# -*- coding: utf-8 -*-
# cython: profile=False, boundscheck=False

# imports ###################################################################

import cython

from cython.parallel import prange
cimport openmp

import numpy as np
cimport numpy as np


# histogram #################################################################

def bootstrap_bounds(object[np.float64_t] ts, N, steps=1):
	"""build initial _t, _lo, and _hi arrays"""
	samples = list(int((N-1)*s/(steps-1)) for s in range(steps))
	t = np.array(ts[samples], dtype=np.float64)
	lo = np.array(samples, dtype=np.int64)
	hi = np.array(samples, dtype=np.int64)
	return t, lo, hi


def update_bounds(object[np.float64_t] ts, N,
                  np.ndarray[np.float64_t] t, np.ndarray[np.float64_t] _t,
                  np.ndarray[np.int64_t] _lo, np.ndarray[np.int64_t] _hi):
	"""update bounds by searching t in old _t"""
	if t[0] < _t[0]:
		lo_0 = _lo[0]
		step = 1
		while True:
			lo_0 -= step
			if lo_0 < 0:
				lo_0 = 0
				break
			if ts[lo_0] <= t[0]:
				break
			step *= 2
		_t[0]  = ts[lo_0]
		_lo[0] = lo_0
	if _t[-1] < t[-1]:
		hi_W = _hi[-1]
		step = 1
		while True:
			hi_W += step
			if hi_W > N-1:
				hi_W = N-1
				break
			if t[-1] <= ts[hi_W]:
				break
			step *= 2
		_t[-1]  = ts[hi_W]
		_hi[-1] = hi_W
	
	cdef np.int32_t W = t.size
	cdef np.int32_t K = _t.size
	cdef np.int32_t l = 0
	cdef np.int32_t h = 0
	cdef np.ndarray[np.int64_t] lo = np.empty(W, dtype=np.int64)
	cdef np.ndarray[np.int64_t] hi = np.empty(W, dtype=np.int64)
	
	for k in range(W):
		while h < K and _t[h] < t[k]:
			h += 1
		while l < K and _t[l] <= t[k]:
			l += 1
		lo[k] = _lo[l-1]
		hi[k] = _hi[h]
	
	return t, lo, hi


def	set_num_threads(n):
	openmp.omp_set_num_threads(n)


@cython.wraparound(False)
def bisect(object[np.float64_t] ts,
           np.ndarray[np.float64_t] t,
           np.ndarray[np.int64_t] lo, np.ndarray[np.int64_t] hi,
           double runtime_ms):
	"""anytime bisecting t in ts using known bounds, using prange"""
	cdef np.int64_t mid
	cdef int i, stop = t.size
	cdef double start = openmp.omp_get_wtime()
	cdef int not_finished = 1
	with nogil:
		while not_finished > 0 and (openmp.omp_get_wtime()-start)*1000. <= runtime_ms:
			not_finished = 0
			for i in prange(stop):
				mid = (lo[i] + hi[i])//2
				if t[i] < ts[mid]:
					hi[i] = mid
				else:
					lo[i] = mid+1
				if lo[i] < hi[i]:
					not_finished += 1
	return np.floor_divide(np.add(lo, hi), 2), np.subtract(hi, lo)
