#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#############################################################################
# imports ###################################################################

import sys
import getopt
import textwrap
import time
import os
import glob

import numpy as np
np.seterr(all='ignore')

import bivis


#############################################################################
# command line ##############################################################

EVENT, DISCRETE, SAMPLED = 'e', 'd', 's'

DEFAULTS = {
	# data
	'data':        EVENT,
	'dtype':       'float64',
	'stddev':      False,
	'carry':       None,
	'N':           None,
	
	# algo tunning
	'B':           10,
	'runtime_max': 15,
	'n_threads':   None,
	
	# display
	'sigma':       4,
	'unit':        's',
	'fmt':         '%Y-%m-%d %H:%M',
	'color':       None,
	'scale':       None,

	# geometry
	'W':           1280,
	'H':            800,
	'X':              0,
	'Y':              0,
	
	# multi process
	'column':      0,
	'port':        None,
	
	# websocket server
	'wserver':     False,
	
	# opengl
	'hidpi':       False,

	# dev tools
	'profiling':   False,
	'manual':      False,
}


def parse_geometry(geom):
	"""(somewhat) parse WxH+X{+|-}Y"""
	size, *origin = geom.split('+', 1)
	if not origin:
		origin = DEFAULTS['X'], DEFAULTS['Y']
	else:
		origin = origin[0]
		if '-' in origin:
			x, y = origin.split('-', 1)
			origin = int(x), -int(y)
		else:
			x, y = origin.split('+', 1)
			origin = int(x), int(y)
	if not size:
		size = DEFAULTS['W'], DEFAULTS['H']
	else:
		size = (int(u) for u in size.split('x', 1))
	return size, origin


def exit_usage(message=None, code=0):
	if message:
		sys.stderr.write('%s\n' % message)
	sys.stderr.write(textwrap.dedent('''\
	Usage: %(name)s [-hd:n:b:t:s:u:f:c:p:mj:] [filename]
		-h  --help                print this help message then exit
		-d  --data <e|d|s>        show event|discrete|sampled data (defaults to %(data)r)
		    --dtype <dt>          use dt dtype to interpret data types (defaults to %(dtype)r)
		    --stddev              enable stddev (provided t, sum v, sum v^2 inputs)
		    --carry <filename>    load filename as overflow values
		-n  --scale <n>           use the first 2**int(n) events (if not set, take the whole data set)
		-b  --bits <b>            use 2**int(b) buckets to bootstrap histogram (defaults to %(B)r)
		-t  --time <ms>           use float(ms) time budget for histogram (defaults to %(runtime_max)r)
		-j  --jobs <t>            use int(t) threads (if not set, does not parallelize bisect)
		-s  --sigma <s>           use int(s) as sigma for gaussian kernel (defaults to %(sigma)r)
		-u  --unit <u>            use u as time unit (defaults to %(unit)r)
		-f  --format <fmt>        use fmt as format string for time (defaults to %(fmt)r, if set empty, do not format the time)
		    --color <name>        use colorbrewer.name color sheme or ...
		            <0xrrggbb …>  ... use listed colors (if not set use error)
		    --max <m>             use float(m) as maximum for vertical axis (if not set, make a guess)
		-g  --geometry <WxH+X+Y>  use int(W), int(H), int(X), int(Y) as window size and position
		-c  --column <c>          use int(c) as column in multiscreen sessions (defaults to %(column)r)
		-p  --port <p>            use int(p) as port for broadcasting interaction)
		-w  --wserver             run as a websocket server (instead of opengl display)
		    --hidpi               enable hidpi
		    --profiling           enable profiling
		-m  --manual              enable manual redisplay
	''') % dict(name=name, **DEFAULTS))
	sys.exit(code)


name, *args = sys.argv
try:
	options, args = getopt.getopt(args, 'hd:n:b:t:j:s:u:f:g:c:p:wm',
	                              ['help',
	                               'data=', 'dtype=', 'stddev', 'carry=', 'scale=',
	                               'bits=', 'time=', 'jobs=',
	                               'sigma=', 'unit=', 'format=', 'color=', 'max=',
	                               'geometry=', 'column=', 'port=', 'wserver', 'hidpi',
	                               'profiling', 'manual'])
except getopt.GetoptError as message:
	exit_usage(message, 1)

if len(args) > 1:
	exit_usage('at most one argument is expected', 1)

data          = DEFAULTS['data']
dtype         = DEFAULTS['dtype']
stddev        = DEFAULTS['stddev']
carry         = DEFAULTS['carry']
N             = DEFAULTS['N']
B             = DEFAULTS['B']
runtime_max   = DEFAULTS['runtime_max']
n_threads     = DEFAULTS['n_threads']
sigma         = DEFAULTS['sigma']
unit          = DEFAULTS['unit']
fmt           = DEFAULTS['fmt']
color         = DEFAULTS['color']
scale         = DEFAULTS['scale']
X             = DEFAULTS['X']
Y             = DEFAULTS['Y']
W             = DEFAULTS['W']
H             = DEFAULTS['H']
column        = DEFAULTS['column']
port          = DEFAULTS['port']
wserver       = DEFAULTS['wserver']
hidpi         = DEFAULTS['hidpi']
profiling     = DEFAULTS['profiling']
manual        = DEFAULTS['manual']



for opt, value in options:
	if opt in ['-h', '--help']:
		exit_usage()
	elif opt in ['-d', '--data']:
		data = value[0]
	elif opt in ['--dtype']:
		dtype = value
	elif opt in ['--stddev']:
		stddev = True
	elif opt in ['--carry']:
		carry = value
	elif opt in ['-n', '--scale']:
		N = 2**int(value)
	elif opt in ['-b', '--bits']:
		B = int(value)
	elif opt in ['-t', '--time']:
		runtime_max = float(value)
	elif opt in ['-j', '--jobs']:
		n_threads = int(value)
	elif opt in ['-s', '--sigma']:
		sigma = int(value)
	elif opt in ['-u', '--unit']:
		unit = value
	elif opt in ['-f', '--format']:
		fmt = value
	elif opt in ['--color']:
		color = value
	elif opt in ['--max']:
		scale = 1./float(value)
	elif opt in ['-g', '--geometry']:
		(W, H), (X, Y) = parse_geometry(value)
	elif opt in ['-c', '--column']:
		column = int(value)
	elif opt in ['-p', '--port']:
		port = int(value)
	elif opt in ['-w', '--wserver']:
		wserver = True
	elif opt in ['--hidpi']:
		hidpi = True
	elif opt in ['--profiling']:
		profiling = True
	elif opt in ['-m', '--manual']:
		manual = True


#############################################################################
# profiling #################################################################

if profiling:
	import cProfile, pstats, atexit
	pr = cProfile.Profile()
	
	@atexit.register
	def report():
		ps = pstats.Stats(pr).sort_stats('tottime')
		ps.print_stats()

	pr.enable()


#############################################################################
# data ######################################################################

if color is not None:
	colors = color.split()
	if len(colors) > 1:
		def rgb(h):
			"r, g, b for hexa 0xRRGGBB"
			h = int(h, 16)
			r = (h >> 16) & 0xff
			g = (h >>  8) & 0xff
			b = (h >>  0) & 0xff
			return r/256, g/256, b/256
		colors = [rgb(c) for c in colors]
	else:
		from utils import colorbrewer
		colorset = getattr(colorbrewer, color)
		colors = colorset[max(colorset)]

if carry is not None:
	carry = np.genfromtxt(carry)
	
if len(args) == 1:
	filename = args[0]
	base, ext = os.path.splitext(filename)
	assert ext == '.dat'
	ts = np.memmap(filename, dtype=np.float64, mode='r')
	filenames = glob.glob(base + '.*' + ext)
	svs = [
		np.memmap(filename, dtype=np.dtype(dtype), mode='r')
		for filename in sorted(filenames)
	]
else:
	d = np.genfromtxt(sys.stdin.buffer, skip_header=1)
	try:
		ts = d[:,0]
		_, k = d.shape
	except IndexError:
		ts = d
		k = 1
	svs = [
		d[:,i]
		for i in range(1, k)
	]

if	svs == []:
	assert data == 'e'
	svs = [None]

if stddev:
	assert len(svs) % 2 == 0
	sv2s = [svs[2*i+1] for i in range(len(svs)//2)]
	svs  = [svs[2*i]   for i in range(len(svs)//2)]

if data == 'e':
	def aggregates(sv, ts, t, i):
		return i
elif data == 'd':
	def aggregates(sv, ts, t, i):
		return sv[i]
elif data == 's':
	def aggregates(sv, ts, t, i):
		i1 = i+1
		t_0 = ts[i]
		t_1 = ts[i1]
		a = (t-t_0)/(t_1-t_0)
		return sv[i]*(1.-a) + sv[i1]*a

if N is None:
	N = len(ts)
else:
	N = min(N, len(ts))

print('loaded %s events (10**%s)' % (format(N, ',d'), format(np.log10(N), '.2f')))

def bootstrap():
	global t, lo, hi
	t, lo, hi = bivis.bootstrap_bounds(ts, N, steps=2**B)
bootstrap()

if n_threads:
	bivis.set_num_threads(n_threads)


#############################################################################
# bisection #################################################################

k = 3 # how many sigma we keep outside the view

dt = (ts[N-2]-ts[1])/W
t_min, t_max = ts[1]-2*k*sigma*dt, ts[N-2]+2*k*sigma*dt


# screen/data mapping

def compute_borders(t_min, t_max, W):
	borders = np.linspace(t_min, t_max, W)
	return borders


# data

def step_bivis(draw_frame, post_redisplay):
	global t, lo, hi
	t, lo, hi = bivis.update_bounds(ts, N,
	                                compute_borders(t_min, t_max, W+2*k*sigma+1),
	                                t, lo, hi)
	mid, width = bivis.bisect(ts, t, lo, hi, runtime_max)
	draw_frame(t, mid, width)
	lo = mid - width//2
	hi = lo + width
	
	if np.any(width > 0.) and not manual:
		post_redisplay()


#############################################################################
# interaction ###############################################################

def pan(dx, then=None):
	global t_min, t_max
	dt = (t_max-t_min)/(W+2*k*sigma)
	t_min -= dx*dt
	t_max -= dx*dt
	if then: then()

def zoom(x, du, then=None):
	global t_min, t_max
	dt = (t_max-t_min)/(W+2*k*sigma)
	ds = np.exp(du*.01)
	t_min += (1-ds)*(dt*(x+k*sigma))
	t_max = t_min + dt*(W+2*k*sigma)*ds
	if then: then()

opacity = 1.

def key(c, then=None):
	if c == b'q':
		sys.exit(0)
	elif c == b'b': # clear bisect cache
		bootstrap()
	elif c in b'[]':
		global sigma, t_min, t_max
		dt = (t_max-t_min)/(W+2*k*sigma)
		if c == b'[':
			if sigma <= 1:
				return
			sigma -= 1
			t_min += k*dt
			t_max -= k*dt
		else:
			sigma += 1
			t_min -= k*dt
			t_max += k*dt
	elif c in b'()':
		global opacity
		if c == b'(':
			opacity = max(0., opacity-.05)
		else:
			opacity = min(1., opacity+.05)
	elif c in b',.': # zoom
		if c == b',':
			zoom(x, -15)
		else:
			zoom(x, 15)
	elif c in b'<>': # pan
		if c == b'<':
			pan(-100)
		else:
			pan(100)
	else:
		return False
	if then: then()
	return True


#############################################################################
# visualization #############################################################

if wserver: # websocket server version ######################################
	
	import json
	import asyncio as asio
	import websockets as ws
	
	async def reader(s, ready):
		x0 = None
		async for msg in s:
			match json.loads(msg):
				case {'evt': 'reshape', 'width': width}:
					global W
					W = width
				case {'evt': 'mousedown', 'b': 1, 'x': x}:
					x0 = x
				case {'evt': 'mousemove', 'x': x1}:
					if x0 is None: continue
					pan(x1 - x0)
					x0 = x1
				case {'evt': 'mouseup'}:
					x0 = None
				case {'evt': 'wheel', 'x': x, 'w': w}:
					zoom(x, w)
			ready.set()
	
	async def writer(s, ready):
		ys = None
		def draw_frame(t, mid, width):
			# simple version, only handling first column, no carry, etc.
			nonlocal ys
			a = aggregates(svs[0], ts, t, mid-1)
			d = np.diff(a)
			ys = np.nan_to_num(d)/np.diff(t)
		
		while True:
			await ready.wait()
			ready.clear()
			step_bivis(draw_frame, ready.set)
			await s.send(ys[k*sigma:W+k*sigma].tobytes())
	
	async def loop(s):
		ready = asio.Event()
		await asio.gather(reader(s, ready), writer(s, ready))
	
	async def main():
		async with ws.serve(loop, 'localhost', '5678'):
			await asio.Future()
	
	try:
		asio.run(main())
	except KeyboardInterrupt:
		print('bye!')
		raise SystemExit


else: # OpenGL version ######################################################

	import OpenGL
	OpenGL.ERROR_CHECKING = False
	OpenGL.ERROR_LOGGING = False
	OpenGL.ERROR_ON_COPY = True
	OpenGL.STORE_POINTERS = False

	from OpenGL.GL import *
	from OpenGL.GLUT import *

	GLUT_WHEEL_UP    = 3
	GLUT_WHEEL_DOWN  = 4
	GLUT_WHEEL_LEFT  = 5
	GLUT_WHEEL_RIGHT = 6

	from utils import text
	from utils import filter


	y_offset = 0.
	z_offset = H/len(svs)
	stack = 1.

	def draw_frame(t, mid, width):
		errors = width[1:]+width[:-1]
		hues = 1./np.log2(errors[k*sigma:W+k*sigma]+2)
		y0s = np.zeros(W)
	
		if carry is not None:
			places = np.searchsorted(t, carry[:,0])
	
		glPushMatrix()
		for i, sv in enumerate(svs):
			a = aggregates(sv, ts, t, mid-1)
			d = np.diff(a)
			if carry is not None:
				for p, c in zip(places, carry[:,i+1]):
					if p-1 < 0:
						continue
					elif p-1 >= len(d):
						break
					d[p-1] += c
			ys = np.nan_to_num(d)/np.diff(t)
		
			fys = filter.box_1d(ys, sigma)[k*sigma:W+k*sigma]
			alphas = 1./(1+fys/ys[k*sigma:W+k*sigma])
			alphas *= (1.-opacity)
			alphas += opacity
		
			global scale
			if scale == None:
				scale = .8*H/np.max(fys)/len(svs)
		
			glPushMatrix()
			glTranslate(0., y_offset, 0.)
			glScale(1, scale, scale)
		
			y1s = y0s + fys

			glColor(0., 0., 0.)
			for yis in [y0s, y1s]:
				glBegin(GL_LINE_STRIP)
				for x, (yi, y) in enumerate(zip(yis, fys)):
					glVertex(x, yi, -y)
				glEnd()

			glBegin(GL_LINES)
			if color:
				r, g, b = colors[i % len(colors)]
				for x, (y0, y1, y, a) in enumerate(zip(y0s, y1s, fys, alphas)):
					glColor(r, g, b, a)
					glVertex(x, y0, -y)
					glVertex(x, y1, -y)
			else:
				for x, (y0, y1, y, r, g, a) in enumerate(zip(y0s, y1s, fys, 1-hues, (1+hues)/2, alphas)):
					glColor(r, g, .2, a)
					glVertex(x, y0, -y)
					glVertex(x, y1, -y)
			glEnd()
		
			if stddev:
				sv2 = sv2s[i]
				a2 = aggregates(sv2, ts, t, mid-1)
				y2s = np.diff(a2)/np.diff(t)
				ss = np.nan_to_num(np.sqrt(y2s - np.square(ys)))
			
				glColor(0., 0., 0., .5)
				glBegin(GL_TRIANGLE_STRIP)
				for x, (y, s) in enumerate(zip(y1s, ss[k*sigma:W+k*sigma])):
					glVertex(x, y-s, -1)
					glVertex(x, y+s, -1)
	#			for x, (y0, y, s) in enumerate(zip(y0s, ys[k*sigma:W+k*sigma], ss[k*sigma:W+k*sigma])):
	#				glVertex(x, y0+y-s, -1)
	#				glVertex(x, y0+y+s, -1)
				glEnd()
			
			
			glPopMatrix()
		
			y0s += stack*fys
			glTranslate(0., (1.-stack)*z_offset, -(1.-stack)*z_offset)
	
		glPopMatrix()


	def draw_scales():
		dt = (t_max-t_min)/(W+2*k*sigma)

		# time scale
		glColor(0, 0, 0, .25)
		glBegin(GL_LINES)
		du = 2 ** np.floor(np.log2(W*dt/5))
		for i in range(int(np.floor((t_min+k*sigma*dt)/du)), int(np.ceil((t_max-k*sigma*dt)/du))):
			t = i*du
			x = (t-t_min)/dt-k*sigma
			glVertex(x, 0)
			glVertex(x, H)
		glEnd()
		glColor(0, 0, 0)
		glBegin(GL_LINE_STRIP)
		glVertex(W-66, H-21)
		glVertex(W-66, H-23)
		glVertex(W-66-sigma, H-23)
		glVertex(W-66-sigma, H-21)
		glEnd()
		text.display('%.2e%s' % (sigma*dt, unit), W-60, H-26)
	
		# density scale
		glPushMatrix()
		glTranslate(0., y_offset, 0.)
		glScale(1, scale, 1)
		glColor(0, 0, 0, .25)
		glBegin(GL_LINES)
		for i in range(-10, 25):
			v = 2**i
			if data != 's':
				v /= dt
			glVertex(0, v)
			glVertex(W, v)
		glEnd()
		for i in range(-10, 25):
			v = 2**i
			if data != 's':
				v /= dt
			glColor(0, 0, 0, min(v, .5))
			text.display('%.2f' % v, 4, v)
		glPopMatrix()
	
		# cursor
		glColor(0, 0, 0, .5)
		glBegin(GL_LINES)
		glVertex(x0-1, 0)
		glVertex(x0-1, H)
		glVertex(0, H-y0)
		glVertex(W, H-y0)
		glEnd()
		glColor(0, 0, 0)
		t = t_min + (x0+k*sigma)*dt
		if fmt:
			t = time.localtime(t)
			t = time.strftime(fmt, t)
		else:
			t = str(t)
		text.display(t, x0+4, H-16)
		s = (H-y0-y_offset)/scale
		text.display('%.2f' % s, 4, H-y0-16)


	# opengl

	def init_gl():
		glClearColor(1, 1, 1, 1)
		glColor(0, 0, 0)
		glEnable(GL_BLEND)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		glEnable(GL_DEPTH_TEST)
		text.init()

	def display():
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
		step_bivis(draw_frame, glutPostRedisplay)
		draw_scales()
		glutSwapBuffers()

	def reshape(width, height):
		global W, H
		W, H = width, height
		glViewport(0, 0, width, height)
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		glOrtho(0, width, 0, height, 0, 100000)
		glTranslate(.5, .5, 0.)
		glMatrixMode(GL_MODELVIEW)
		glLoadIdentity()


	#############################################################################
	# interaction ###############################################################

	if port != None:
		import socket
		import select
	
		interaction_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
		interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
		interaction_socket.bind(('', port))

		MESSAGE = '%(id)s %(width)s %(column)s %(t_min)s %(t_max)s\n'
		_id = str(os.getpid())

		def broadcast():
			dt = (t_max-t_min)/(W+2*k*sigma)
			interaction_socket.sendto((MESSAGE % {
				'id': _id,
				'width': W,
				'column': column,
				't_min': t_min + k*sigma*dt,
				't_max': t_max - k*sigma*dt,
			}).encode(), ('<broadcast>', port))

	
		def idle():
			msg = b''
			while True:
				r, _, _ = select.select([interaction_socket], [], [], 1./200)
				if r == []:
					break
				msg = interaction_socket.recv(1024)
			try:
				msg = msg.decode()
				msg = msg.split('\n')[-2]
			except IndexError:
				return
			other_id, other_W, other_column, other_t_min, other_t_max = msg.split()
			if other_id == _id:
				return
			other_W, other_column = int(other_W), int(other_column)
			other_t_min, other_t_max = float(other_t_min), float(other_t_max)
			dt = (other_t_max-other_t_min)/other_W
			dT = (column-other_column)*(other_t_max-other_t_min)
		
			global t_min, t_max
			t_min = other_t_min+dT-k*sigma*dt
			t_max = other_t_max+dT+k*sigma*dt
			glutPostRedisplay()
			if other_W != W:
				glutReshapeWindow(other_W, H)
	else:
		def broadcast(): pass
		idle = None

	_shot = 0
	def screen_shot(name="screen_shot.%03i.png"):
		"""window screenshot."""
		width, height = glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT)
		data = glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE)
	
		import png
		global _shot
		png.write(open(name % _shot, "wb"), width, height, 3, data)
		_shot += 1


	WINDOW_SIZE = None

	def keyboard(c, x, y):
		if key(c):
			pass
		elif c == b'f':
			global WINDOW_SIZE
			if WINDOW_SIZE is None:
				WINDOW_SIZE = glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT)
				glutFullScreen()
			else:
				glutReshapeWindow(*WINDOW_SIZE)
				WINDOW_SIZE = None
		elif c == b's':
			screen_shot()
		elif c in b'{}':
			global scale
			if c == b'{':
				scale /= 1.05
			else:
				scale *= 1.05
		elif c in b'-=':
			global stack
			if c == b'-':
				stack = min(1., stack+.05)
			else:
				stack = max(0., stack-.05)
		elif c in b'_+':
			global z_offset
			if c == b'_':
				z_offset = max(0., z_offset-1)
			else:
				z_offset += 1
		elif c == b'r': # redisplay
			pass
		glutPostRedisplay()



	
	IDLE, PANNING, ZOOMING, ORTHOZOOMING = range(4)
	interaction = IDLE
	vertical_pan = None
	x0, y0 = 0, 0

	def mouse(button, state, x, y):
		global interaction, xo, yo, x0, y0, vertical_pan
		modifiers = glutGetModifiers()
		if state == GLUT_DOWN:
			if button == GLUT_LEFT_BUTTON:
				interaction = PANNING
				vertical_pan = None
			elif button == GLUT_RIGHT_BUTTON:
				interaction = ZOOMING
			elif button == GLUT_MIDDLE_BUTTON:
				interaction = ORTHOZOOMING
			elif button in (GLUT_WHEEL_UP, GLUT_WHEEL_DOWN):
				if modifiers == 0:
					zoom(x, -1 if button == GLUT_WHEEL_DOWN else 1, broadcast)
			elif button in (GLUT_WHEEL_LEFT, GLUT_WHEEL_RIGHT):
				if modifiers == 0:
					pan(-5 if button == GLUT_WHEEL_RIGHT else 5, broadcast)
		else:
			if button in (GLUT_WHEEL_UP, GLUT_WHEEL_DOWN, GLUT_WHEEL_LEFT, GLUT_WHEEL_RIGHT):
				if modifiers & GLUT_ACTIVE_SHIFT:
					keyboard(b'{' if button in (GLUT_WHEEL_DOWN, GLUT_WHEEL_RIGHT) else b'}', x, y)
				elif modifiers & GLUT_ACTIVE_ALT:
					keyboard(b'[' if button in (GLUT_WHEEL_DOWN, GLUT_WHEEL_RIGHT) else b']', x, y)
			interaction = IDLE
		xo, yo = x0, y0 = x, y
		glutPostRedisplay()

	def motion(x1, y1):
		global xo, yo, x0, y0, vertical_pan
		dx, dy = x1-x0, y1-y0
		if interaction == PANNING:
			global vertical_pan, y_offset
			if vertical_pan is None:
				vertical_pan = abs(dy) > abs(dx)
			if vertical_pan:
				y_offset -= dy
			pan(dx, broadcast)
		elif interaction == ZOOMING:
			zoom(xo-1, dx-dy, broadcast)
		elif interaction == ORTHOZOOMING:
			pan(-dx)
			zoom(x1-1, -dy, broadcast)
		x0, y0 = x1, y1
		glutPostRedisplay()


	#############################################################################
	# main ######################################################################

	glutInit(sys.argv)
	display_string = [b'rgba', b'double', b'depth', b'samples']
	if hidpi:
		display_string.append(b'hidpi')
	glutInitDisplayString(b' '.join(display_string))
	glutInitWindowSize(W, H)
	glutInitWindowPosition(X, Y)
	glutCreateWindow(name)

	glutDisplayFunc(display)
	glutReshapeFunc(reshape)
	glutKeyboardFunc(keyboard)
	glutMouseFunc(mouse)
	glutMotionFunc(motion)
	glutPassiveMotionFunc(motion)
	glutIdleFunc(idle)

	init_gl()

	glutMainLoop()

