#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#############################################################################
# imports ###################################################################

import sys
import getopt
import textwrap
import time
import os

from math import exp, log

import numpy as np
np.seterr(all='ignore')
import scipy.ndimage as ni
import pyopencl as cl

import OpenGL
OpenGL.ERROR_CHECKING = False
OpenGL.ERROR_LOGGING = False
OpenGL.ERROR_ON_COPY = True
OpenGL.STORE_POINTERS = False

from OpenGL.GLUT import *
from OpenGL.GL import *

GLUT_WHEEL_UP   = 3
GLUT_WHEEL_DOWN = 4

import bivis
try:
	from pargsort import pargsort
except ImportError:
	# non parallel fallback
	def pargsort(u): return np.argsort(u, kind='mergesort')


#############################################################################
# command line ##############################################################

DEFAULTS = {
	'B':           10,
	'runtime_max': 50,
	'sigma':       1.,
	'gamma':       .5,
	'max_count':   None,
	'column':      0,
	'row':         0,
	'port':        None,
	'profiling':   False,
	'manual':      False,
	'n_threads':   None,
	'hidpi':       False,
}

def exit_usage(message=None, code=0):
	if message:
		sys.stderr.write('%s\n' % message)
	sys.stderr.write(textwrap.dedent('''\
	Usage: %(name)s [-hb:t:s:g:m:c:r:p:j:] [filename]
		-h  --help             print this help message then exit
		-b  --bits <b>         use 2**int(b) buckets to bootstrap histogram (defaults to %(B)r)
		-t  --time <ms>        use float(ms) time budget for histogram (defaults to %(runtime_max)r)
		-s  --sigma <s>        use int(s) as sigma for gaussian kernel (defaults to %(sigma)r)
		-g  --gamma <g>        use g for 'gamma correction' (defaults to %(gamma)r)
		-m  --max <m>          use 10**m for max_count (defaults to Nxln(N))
		-c  --column <c>       use int(c) as column in multiscreen sessions (defaults to %(column)r)
		-r  --row <r>          use int(r) as row in multiscreen sessions (defaults to %(row)r)
		-p  --port <p>         use int(p) as port for broadcasting interaction)
		    --profiling        enable profiling
		    --manual           enable manual redisplay
		    --hidpi            enable hidpi
		-j  --jobs <t>         use int(t) threads (if not set, does not parallelize bisect)
	''') % dict(name=name, **DEFAULTS))
	sys.exit(code)


name, *args = sys.argv
try:
	options, args = getopt.getopt(args, 'hb:t:s:g:m:c:r:p:j:',
	                              ['help',
	                               'bits=', 'time=', 'sigma=',
	                               'gamma=', 'max=',
	                               'column=', 'row=', 'port=', 'jobs=',
	                               'profiling', 'manual', 'hidpi'])
except getopt.GetoptError as message:
	exit_usage(message, 1)

if len(args) > 1:
	exit_usage('at most one argument is expected', 1)

B             = DEFAULTS['B']
runtime_max   = DEFAULTS['runtime_max']
sigma         = DEFAULTS['sigma']
gamma         = DEFAULTS['gamma']
max_count     = DEFAULTS['max_count']
column        = DEFAULTS['column']
row           = DEFAULTS['row']
port          = DEFAULTS['port']
profiling     = DEFAULTS['profiling']
manual        = DEFAULTS['manual']
n_threads     = DEFAULTS['n_threads']
hidpi         = DEFAULTS['hidpi']

for opt, value in options:
	if opt in ['-h', '--help']:
		exit_usage()
	elif opt in ['-b', '--bits']:
		B = int(value)
	elif opt in ['-t', '--time']:
		runtime_max = float(value)
	elif opt in ['-s', '--sigma']:
		sigma = int(value)
	elif opt in ['-g', '--gamma']:
		gamma = float(value)
	elif opt in ['-m', '--max']:
		max_count = 10**float(value)
	elif opt in ['-c', '--column']:
		column = int(value)
	elif opt in ['-r', '--row']:
		row = int(value)
	elif opt in ['-p', '--port']:
		port = int(value)
	elif opt in ['--profiling']:
		profiling = True
	elif opt in ['--manual']:
		manual = True
	elif opt in ['-j', '--jobs']:
		n_threads = int(value)
	elif opt in ['--hidpi']:
		hidpi = True


#############################################################################
# profiling #################################################################

if profiling:
	import cProfile, pstats, atexit
	pr = cProfile.Profile()
	
	@atexit.register
	def report():
		ps = pstats.Stats(pr).sort_stats('tottime')
		ps.print_stats()

	pr.enable()


#############################################################################
# data ######################################################################

if len(args) == 1:
	filename = args[0]
	base, ext = os.path.splitext(filename)
	if ext == '.h5':
		import tables
		h5file = tables.open_file(filename)
		cities = h5file.root.ts[:]
		try:
			cumpop = h5file.root.ss[:]
		except:
			pass
	elif ext == '.dat':
		cities = np.memmap(filename, dtype=np.float64, mode='r')
		try:
			cumpop = np.memmap(base+'.p'+ext, dtype=np.int64, mode='r')
		except:
			pass
else:
	data = np.genfromtxt(sys.stdin.buffer, skip_header=1)
	try:
		cities = data[:,0]
	except IndexError:
		cities = data
	try:
		cumpop = data[:,4]
	except IndexError:
		pass

try:
	cumpop
	def values(mid):
		return cumpop[mid]
except NameError:
	def values(mid):
		return mid

N = len(cities)

if max_count is None:
	max_count = values(N-1)*np.log(values(N-1)) # guess some value

print('loaded %s points (10**%s)' % (format(N, ',d'), format(np.log10(N), '.2f')))

def bootstrap():
	global p, lo, hi
	p, lo, hi = bivis.bootstrap_bounds(cities, N, steps=2**B)
bootstrap()

if n_threads:
	bivis.set_num_threads(n_threads)


#############################################################################
# processing ################################################################

BITS = 12

ctx = cl.Context(dev_type=cl.device_type.GPU)
ctx = cl.create_some_context()
queue = cl.CommandQueue(ctx)
hilbert = cl.Program(ctx, """
float hilbert(float *a, float *o);
float hilbert(float *a, float *o) {
	float u = 0.f;
	float du = 1.f;
	float na, no;
	for(int l=0; l < %(BITS)s; l++) {
		du /= 4.f;
		na = *a*2.f; if(na >= 1.f) na -= 1.f;
		no = *o*2.f; if(no >= 1.f) no -= 1.f;
		if(*o < .5f) {
			if(*a < .5f) {
				*a = no;
				*o = na;
			} else {
				*a = 1.f-no;
				*o = 1.f-na;
				u += 3.f*du;
			}
		} else {
			if(*a < .5f) {
				u += du;
			} else {
				u += 2.f*du;
			}
			*a = na;
			*o = no;
		}
	}
	return u;
}

__kernel void hash(
	const float x, const float y,
	const float sx, const float sy,
	__global float *HU, __global float *LU) {
	
	int H = get_global_size(0);
	int W = get_global_size(1);
	int i = get_global_id(0);
	int j = get_global_id(1);
	float a = 1.f*i/H/sy-y; // lat
	while(a < 0.f)  { a += 1.f; }
	while(a >= 1.f) { a -= 1.f; }
	float o = 1.f*j/W/sx-x; // long
	while(o < 0.f)  { o += 1.f; }
	while(o >= 1.f) { o -= 1.f; }
	HU[i*W+j] = hilbert(&a, &o);
	LU[i*W+j] = hilbert(&a, &o);
}

__kernel void fish_hash(
	const float x, const float y,
	const float sx, const float sy,
	const float CJ, const float CI,
	__global float *HU, __global float *LU) {
	const float R = 100;
	
	int H = get_global_size(0);
	int W = get_global_size(1);
	int i = get_global_id(0);
	int j = get_global_id(1);
	float CY = H-CI;
	float r = hypot(i-CY, j-CJ);
	float alpha = atan2(i-CY, j-CJ);
	float m = r/R;
	if(m < 2) {
		m /= 4.f;
	} else if(m < 3) {
		m = (m-2)*2.5f + .5f;
	}
	float X = m*R*cos(alpha) + CJ;
	float Y = m*R*sin(alpha) + CY;
	
	float a = Y/H/sy-y; // lat
	float o = X/W/sx-x; // long
	
//	float a = 1.f*i/H/sy-y; // lat
//	float o = 1.f*j/W/sx-x; // long

	while(a < 0.f)  { a += 1.f; }
	while(a >= 1.f) { a -= 1.f; }
	while(o < 0.f)  { o += 1.f; }
	while(o >= 1.f) { o -= 1.f; }
	HU[i*W+j] = hilbert(&a, &o);
	LU[i*W+j] = hilbert(&a, &o);
}
""" % {'BITS':BITS}).build()
hilbert_hash = hilbert.hash
hilbert_hash.set_scalar_arg_dtypes([np.float32, np.float32,
                                    np.float32, np.float32,
                                    None, None])
hilbert_fishhash = hilbert.fish_hash
hilbert_fishhash.set_scalar_arg_dtypes([np.float32, np.float32,
                                        np.float32, np.float32,
                                        np.float32, np.float32,
                                        None, None])


def rescale(ds=None):
	global SCALE, WS, HS, lu, hu, lu_g, hu_g
	if ds is None:
		SCALE = 1
	else:
		scale = SCALE
		SCALE *= ds
		if SCALE < 1:
			SCALE = 1
		if SCALE == scale:
			return
	glPixelZoom(SCALE, SCALE)
	WS, HS = int(W/SCALE), int(H/SCALE)
	hu = np.empty((HS, WS), dtype=np.float32)
	hu_g = cl.Buffer(ctx, cl.mem_flags.WRITE_ONLY, hu.nbytes)
	lu = np.empty((HS, WS), dtype=np.float32)
	lu_g = cl.Buffer(ctx, cl.mem_flags.WRITE_ONLY, lu.nbytes)


_view = None
def hash():
	"""hash pixels"""
	global _view, _r
	
	x, y = center
	if fisheye:
		view = hu.shape, None, x, y, scale, x0, y0
		if view == _view:
			return _r
		
		hilbert_fishhash(queue, hu.shape, None,
		             x, y, scale*(W/WS/SCALE), scale*(H/HS/SCALE), x0, y0,
		             hu_g, lu_g)
	else:
		view = hu.shape, None, x, y, scale
		if view == _view:
			return _r
		hilbert_hash(queue, hu.shape, None,
		             x, y, scale*(W/WS/SCALE), scale*(H/HS/SCALE),
		             hu_g, lu_g)
	cl.enqueue_copy(queue, hu, hu_g)
	cl.enqueue_copy(queue, lu, lu_g)
	u = hu.astype(np.float64)
	u += lu/(4**BITS)
	u.shape = (WS*HS)
	
	o = pargsort(u) # permutation to put cities in order
	p = np.empty(o.size+1, dtype=np.float64)
	p[0], p[-1] = u[o[0]], u[o[-1]]  # endpoints
	p[1:-1] = (u[o[1:]]+u[o[:-1]])/2 # midpoints
	
	i = np.empty(o.size, dtype=np.int64)
	np.put(i, o, np.arange(i.size, dtype=np.int64)) # i[o[k]] = k, i.e. reverse permutation
	
	result = u, p, i
	
	_view = view
	_r = result
	
	return result


def filter(count):
	"""filter counts."""
	count[count<0] = 0.
	count[np.isnan(count)] = 0.
	count *= sigma*sigma
	count /= max_count*scale
	count = np.power(count, gamma)
	count = ni.filters.gaussian_filter(count, sigma)
	return count


#############################################################################
# display ###################################################################

W, H = 1920, 1200
W, H = 1280, 800
W, H = 1024, 768
W, H = 800, 600
W, H = 640, 480
SCALE = 1

WINDOW_SIZE = None


def reshape(width, height):
	"""window reshape callback."""
	global W, H
	W, H = width, height
	glViewport(0, 0, W, H)
	rescale()

NO, HIGH, LOW = range(3)
display_hilbert = NO

def display():
	"""window redisplay callback."""
	u, n_p, i = hash()
	
	if display_hilbert != NO:
		glDrawPixels(WS, HS, GL_LUMINANCE, GL_FLOAT, hu if display_hilbert == HIGH else lu)
		glutSwapBuffers()
		return
	
	global p, lo, hi
	p, lo, hi = bivis.update_bounds(cities, N,
	                                n_p,
	                                p, lo, hi)
	mid, width = bivis.bisect(cities, p, lo, hi, runtime_max)
	draw_frame(p, mid, width, i)
	lo = mid - width//2
	hi = lo + width
	
	if (width > 0).any() and not manual:
		glutPostRedisplay()


def draw_frame(p, mid, width, i):
	glClear(GL_COLOR_BUFFER_BIT)
	
	count = np.divide(np.diff(values(mid)), np.diff(p), dtype=np.float32)
	errors = width[1:]+width[:-1]
	
	count = np.reshape(count[i], (HS, WS)) # put back in order and shape
	count = filter(count)
	glDrawPixels(WS, HS, GL_LUMINANCE, GL_FLOAT, 1.-count)
	
	errors = np.reshape(errors[i], (HS, WS))
	alphas = 1-5/np.log2(errors*SCALE*SCALE+2, dtype=np.float32)
	if False: # bias != 0. makes the program crash on my new laptop when SCALE != 1 ?!
		glPixelTransferf(GL_RED_BIAS, 1.)
		glDrawPixels(WS, HS, GL_ALPHA, GL_FLOAT, alphas)
		glPixelTransferf(GL_RED_BIAS, 0.)
	else:
		overlay = np.zeros((HS, WS, 4), dtype=np.float32)
		overlay[:,:,0] = 1.
		overlay[:,:,3] = alphas
		glDrawPixels(WS, HS, GL_RGBA, GL_FLOAT, overlay)

	glutSwapBuffers()


#############################################################################
# interaction ###############################################################

_shot = 0
def screen_shot(name="screen_shot.%03i.png"):
	"""window screenshot."""
	width, height = glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT)
	data = glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE)
	
	import png
	global _shot
	png.write(open(name % _shot, "wb"), width, height, 3, data)
	_shot += 1


center = 0., 0.
scale = 1.

if port != None:
	import socket
	import select
	
	interaction_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
	interaction_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
	interaction_socket.bind(('', port))

	MESSAGE = '%(id)s %(width)s %(height)s %(column)s %(row)s %(x_min)s %(x_max)s %(y_min)s %(y_max)s\n'
	_id = str(os.getpid())

	def broadcast():
		x, y = center
		interaction_socket.sendto((MESSAGE % {
			'id': _id,
			'width': W,
			'height': H,
			'column': column,
			'row': row,
			'x_min': x,
			'x_max': x+1./scale,
			'y_min': y,
			'y_max': y+1./scale,
		}).encode(), ('<broadcast>', port))
	
	def idle():
		msg = b''
		while True:
			r, _, _ = select.select([interaction_socket], [], [], 1./200)
			if r == []:
				break
			msg = interaction_socket.recv(1024)
		try:
			msg = msg.decode()
			msg = msg.split('\n')[-2]
		except IndexError:
			return
		(other_id, other_W, other_H, other_column, other_row,
		 other_x_min, other_x_max, other_y_min, other_y_max) = msg.split()
		if other_id == _id:
			return
		other_W, other_H, other_column, other_row = int(other_W), int(other_H), int(other_column), int(other_row)
		other_x_min, other_x_max = float(other_x_min), float(other_x_max)
		other_y_min, other_y_max = float(other_y_min), float(other_y_max)
		global center, scale
		scale = 1./(other_x_max-other_x_min)
		x_min = other_x_min+(column-other_column)/scale
		y_min = other_y_min+(row-other_row)/scale
		center = x_min, y_min
		glutPostRedisplay()
#		if (other_W, other_H) != (W, H):
#			glutReshapeWindow(other_W, other_H)
else:
	def broadcast(): pass
	idle = None


fisheye = False

def keyboard(c, x, y):
	"""keyboard callback."""
	if c in [b'q', 27]:
		sys.exit(0)
	
	elif c == b' ':
		global fisheye
		fisheye = True
		glutSetCursor(GLUT_CURSOR_CROSSHAIR)
		glutPostRedisplay()
	
	elif c in b'[]':
		global sigma
		if c == b'[':
			sigma *= 1.1
		else:
			sigma /= 1.1
		print('sigma', sigma)
		glutPostRedisplay()
	
	elif c in b'{}':
		global max_count
		if c == b'{':
			max_count *= 1.1
		else:
			max_count /= 1.1
		print('max_count', max_count)
		glutPostRedisplay()
	
	elif c in b";'":
		global gamma
		if c == b';':
			gamma *= 1.05
		else:
			gamma /= 1.05
		print('gamma', gamma)
		glutPostRedisplay()
	
	elif c in b'-=':
		if c == b'-':
			rescale(1/1.1)
		else:
			rescale(1.1)
		glutPostRedisplay()
	
	elif c == b's':
		screen_shot()
	
	elif c == b'f':
		global WINDOW_SIZE
		if WINDOW_SIZE is None:
			WINDOW_SIZE = glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT)
			glutFullScreen()
		else:
			glutReshapeWindow(*WINDOW_SIZE)
			WINDOW_SIZE = None
	
	elif c == b'h': # switch hilbert display
		global display_hilbert
		display_hilbert += 1
		if display_hilbert > LOW:
			display_hilbert = NO
		glutPostRedisplay()
	
	elif c == b'r': # redisplay
		glutPostRedisplay()
	elif c == b'b': # clear bisect cache
		bootstrap()
		glutPostRedisplay()
	elif c in b',.': # zoom
		if c == b',':
			zoom(x, y, -15)
		else:
			zoom(x, y, 15)
	elif c in b'<>az': # pan
		if c == b'<':
			pan(-15, 0)
		elif c == b'>':
			pan(15, 0)
		elif c == b'a':
			pan(0, -15)
		else:
			pan(0, 15)

def keyup(c, x, y):
	"""keyboard callback."""
	if c == b' ':
		global fisheye
		fisheye = False
		glutSetCursor(GLUT_CURSOR_LEFT_ARROW)
		if glutGetModifiers() & GLUT_ACTIVE_SHIFT:
			zoom(x0, y0, log(4)*100)
		glutPostRedisplay()

panning = False
zooming = False

def mouse(button, state, x, y):
	global x0, y0, panning
	global xz, yz, zooming
	x0, y0 = xz, yz = x, y
	
	if button == GLUT_LEFT_BUTTON:
		panning = (state == GLUT_DOWN)
	
	elif button == GLUT_RIGHT_BUTTON:
		zooming = (state == GLUT_DOWN)
	
	elif button == GLUT_WHEEL_UP:
		zoom(x, y, 1)
	elif button == GLUT_WHEEL_DOWN:
		zoom(x, y, -1)

x0, y0 = 0, 0

def motion(x1, y1):
	global x0, y0, panning
	global xz, yz, zooming
	dx, dy = x1-x0, y1-y0
	x0, y0 = x1, y1
	
	if panning:
		pan(dx, dy)
	
	elif zooming:
		zoom(xz, yz, dx-dy)
		pass
	
	if fisheye:
		glutPostRedisplay()


def pan(dx, dy):
	global center, scale
	cx, cy = center
	dx *= 1./W/scale
	dy *= -1./H/scale
	center = cx+dx, cy+dy

	broadcast()
	glutPostRedisplay()

def zoom(x, y, s):
	global scale, center
	
	cx, cy = center
	x = x/W/scale
	y = (H-y)/H/scale
	
	ds = exp(s*.01)
	cx += (x-cx)*(1.-ds)
	cy += (y-cy)*(1.-ds)
	center = cx/ds, cy/ds
	scale *= ds
	
	broadcast()
	glutPostRedisplay()


#############################################################################
# setup #####################################################################

glutInit(sys.argv)
display_string = [b'rgba', b'double']
if hidpi:
	display_string.append(b'hidpi')
glutInitDisplayString(b' '.join(display_string))
glutInitWindowSize(W, H)
glutCreateWindow(sys.argv[0])

glutReshapeFunc(reshape)
glutDisplayFunc(display)
glutKeyboardFunc(keyboard)
glutKeyboardUpFunc(keyup)
glutMouseFunc(mouse)
glutMotionFunc(motion)
glutPassiveMotionFunc(motion)
glutIdleFunc(idle)

glutSetKeyRepeat(False)

glEnable(GL_BLEND)
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
glClearColor(1., 1., 1., 1.)

glutMainLoop()
