# -*- coding: utf-8 -*-
# cython: profile=False, boundscheck=False, wraparound=False

# imports ###################################################################

cdef extern from "stdlib.h":
	int mergesort(void *base, size_t nel, size_t width,
	              int (*compar)(const void *, const void *)) nogil

from cython.parallel import prange, threadid
from cython.parallel cimport parallel

cimport openmp

import numpy as np
cimport numpy as np


# parallel argsort ##########################################################

cdef const np.float64_t *data

cdef int compare(const void *v1, const void *v2):
	"""comparison function to turn sort into argsort"""
	cdef const np.int64_t *i1 = <np.int64_t *>v1
	cdef const np.int64_t *i2 = <np.int64_t *>v2
	
	cdef np.float64_t d = data[i1[0]] - data[i2[0]]
	if d < 0.:
		return -1
	elif d == 0.:
		return 0
	else:
		return 1

def pargsort(np.ndarray[np.float64_t] u):
	"""parallel argsort
	
	we split the input in num_threads chunks and sort each chunk in parallel,
	then we mergesort again the global array (this step is hopefully O(N)
	since input is mostly sorted).
	we have to rely on a global data because there is no mergesort_r (as there
	is a qsort_r) so pargsort is not reentrant.
	"""
	
	global data
	data = <const np.float64_t *>(&u[0])
	
	cdef size_t size = u.size
	cdef np.ndarray[np.int64_t] order = np.arange(size, dtype=np.int64)
	cdef size_t width = order.itemsize
	
	cdef int num_threads = openmp.omp_get_max_threads()
	cdef size_t chunk_len = size/num_threads
	
	cdef int thread_id
	with nogil, parallel():
		thread_id = openmp.omp_get_thread_num()
		mergesort(&order[thread_id*chunk_len],
		          chunk_len if thread_id < num_threads-1 else size-thread_id*chunk_len,
		          width, compare)
	
	# and the merge
	mergesort(&order[0], size, width, compare) # rely on mergesort being O(N) for mostly sorted input
	return order
